#!/usr/bin/env python
# coding=utf-8
"""
download.py

Part of neoepiscope
Downloads dependencies according to user input to facilitate installation.

Licensed under the MIT license.

The MIT License (MIT)
Copyright (c) 2018 Mary A. Wood, Austin Nguyen,
                   Abhinav Nellore, and Reid Thompson

Adapted from Rail-RNA, which is copyright (c) 2015 
                    Abhinav Nellore, Leonardo Collado-Torres,
                    Andrew Jaffe, James Morton, Jacob Pritt,
                    José Alquicira-Hernández,
                    Christopher Wilks,
                    Jeffrey T. Leek, and Ben Langmead.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import absolute_import, division, print_function
from six.moves import input
from .version import version_number
from .file_processing import which
import signal
import shutil
import tempfile
import sys
import os
import subprocess
from .transcript import gtf_to_cds
from .transcript import cds_to_tree
from distutils.core import Command

download = {
    "Gencode v27 annotation": [
        "ftp://ftp.sanger.ac.uk/pub/gencode/Gencode_human"
        "/release_27/gencode.v27.annotation.gtf.gz"
    ],
    "Gencode v19 annotation": [
        "ftp://ftp.ebi.ac.uk/pub/databases/gencode/"
        "Gencode_human/release_19/"
        "gencode.v19.annotation.gtf.gz"
    ],
    "Bowtie NCBI GRCh38 index": [
        "ftp://ftp.ccb.jhu.edu/pub/data/bowtie_indexes/GRCh38_no_alt.zip",
        "http://verve.webfactional.com/GRCh38_no_alt.zip",
    ],
    "Bowtie UCSC hg19 index": [
        "ftp://ftp.ccb.jhu.edu/pub/data/bowtie_indexes/hg19.ebwt.zip",
        "http://verve.webfactional.com/hg19.ebwt.zip",
    ],
}


def is_exe(fpath):
    """ Tests whether a file is executable.
        fpath: path to file
        Return value: True iff file exists and is executable.
    """
    return os.path.exists(fpath) and os.access(fpath, os.X_OK)


def remove_temporary_directories(temp_dir_paths):
    """ Deletes temporary directory.

        temp_dir_paths: iterable of paths of temporary directories

        No return value.
    """
    for temp_dir_path in temp_dir_paths:
        try:
            shutil.rmtree(temp_dir_path, ignore_errors=True)
        except Exception as e:
            # Don't know what's up, but forge on
            pass


def sig_handler(signum, frame):
    """ Helper function for register_cleanup that's called on signal. """
    import sys

    sys.exit(0)


def register_cleanup(handler, *args, **kwargs):
    """ Registers cleanup on normal and signal-induced program termination.

        Executes previously registered handler as well as new handler.

        handler: function to execute on program termination
        args: named arguments of handler
        kwargs includes keyword args of handler as well as:
            signals_to_handle: list of signals to handle, e.g. [signal.SIGTERM,
                signal.SIGHUP]

        No return value.
    """
    if "signals_to_handle" in kwargs:
        signals_to_handle = kwargs["signals_to_handle"]
        del kwargs["signals_to_handle"]
    else:
        signals_to_handle = [signal.SIGTERM, signal.SIGHUP]
    from atexit import register

    register(handler, *args, **kwargs)
    old_handlers = [
        signal.signal(a_signal, sig_handler) for a_signal in signals_to_handle
    ]
    for i, old_handler in enumerate(old_handlers):
        if (old_handler != signal.SIG_DFL) and (old_handler != sig_handler):

            def new_handler(signum, frame):
                try:
                    sig_handler(signum, frame)
                finally:
                    old_handler(signum, frame)

        else:
            new_handler = sig_handler
        signal.signal(signals_to_handle[i], new_handler)


def print_to_screen(message, newline=True, carriage_return=False):
    """ Prints message to stdout as well as stderr if stderr is redirected.

        message: message to print
        newline: True iff newline should be printed
        carriage_return: True iff carriage return should be printed; also
            clears line with ANSI escape code

        No return value.
    """
    full_message = (
        "\x1b[K"
        + message
        + ("\r" if carriage_return else "")
        + ("\n" if newline else "")
    )
    try:
        sys.stderr.write(full_message)
        if sys.stderr.isatty():
            sys.stderr.flush()
        else:
            try:
                # So the user sees it too
                sys.stdout.write(full_message)
                sys.stdout.flush()
            except UnicodeEncodeError:
                sys.stdout.write(
                    unicodedata.normalize("NFKD", full_message).encode(
                        "ascii", "ignore"
                    )
                )
                sys.stdout.flush()
    except UnicodeEncodeError:
        sys.stderr.write(
            unicodedata.normalize("NFKD", full_message).encode("ascii", "ignore")
        )
        sys.stderr.flush()


class NeoepiscopeDownloader(object):
    """ Convenience class for downloading files so neoepiscope is ready to use.

        Init vars
        -------------
        curl_exe: path to cURL executable; if None, use 'curl'
        download_dir: where to put files; if None, use home directory
            + 'neoepiscope.data'
        yes: if True, answer yes to all questions automatically
        print_log_on_error: print log on error
    """

    def __init__(
        self,
        curl_exe=None,
        download_dir=None,
        yes=False,
        print_log_on_error=False,
        version_number=version_number,
    ):
        print_to_screen("""Configuring neoepiscope v{0} ...""".format(version_number))
        self.download_dir = download_dir
        self.curl_exe = curl_exe
        log_dir = tempfile.mkdtemp()
        self.log_file = os.path.join(log_dir, "neoepiscope_config.err")
        self.log_stream = open(self.log_file, "w")
        self.finished = False
        register_cleanup(remove_temporary_directories, [log_dir])
        self.print_log_on_error = print_log_on_error
        self.yes = yes

    def __enter__(self):
        return self

    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def _print_to_screen_and_log(self, message, **kwargs):
        print(message, file=self.log_stream)
        print_to_screen(message, **kwargs)

    def _bail(self):
        """ Copy log to some temporary dir and GTFO. """
        new_log_file = os.path.join(tempfile.mkdtemp(), "neoepiscope_config.err")
        shutil.copyfile(self.log_file, new_log_file)
        if self.print_log_on_error:
            print_to_screen("Log (also at %s):" % new_log_file)
            print_to_screen("===")
            with open(new_log_file, "r") as log_fh:
                for ln in log_fh:
                    print_to_screen(ln.rstrip())
            print_to_screen("===")
        else:
            print_to_screen("Installation log may be found at %s." % new_log_file)
        sys.exit(1)

    def _yes_no_query(self, question, answer=None):
        """ Gets a yes/no answer from the user if self.yes is not True.

            question: string with question to be printed to console
            answer: boolean that overrides self.yes if an answer should
                be forced

            Return value: boolean
        """
        from distutils.util import strtobool

        if answer is None:
            if self.yes:
                print("\x1b[K%s [y/n]: y" % question)
                return True
        elif answer:
            print("\x1b[K%s [y/n]: y" % question)
            return True
        else:
            print("\x1b[K%s [y/n]: n" % question)
            return False
        while True:
            sys.stdout.write("\x1b[K%s [y/n]: " % question)
            try:
                try:
                    return strtobool(input().lower())
                except KeyboardInterrupt:
                    sys.stdout.write("\n")
                    sys.exit(0)
            except ValueError:
                sys.stdout.write("Please enter 'y' or 'n'.\n")

    def _request_path(self, request, program="", use_which=True):
        """ Gets a path from a user to an installed software

            request: string with request to be printed to console
            use_which: whether to use the which function to verify a path

            Return value: string with response to request
        """
        if use_which:
            which_function = lambda x: which(x)
        else:
            which_function = lambda x: x
        while True:
            sys.stdout.write("%s: " % request)
            try:
                path = which_function(input())
                if path is not None:
                    return path
                else:
                    raise ValueError("".join(["Not a valid install of ", program]))
            except KeyboardInterrupt:
                sys.stdout.write("\n")
                sys.exit(0)

    def check_exe(self, program):
        """ Tests whether an executable is in PATH.
            program: executable to search for
            Return value: path to executable or None if the executable is not
                found.
        """

        def ext_candidates(fpath):
            yield fpath
            for ext in os.environ.get("PATHEXT", "").split(os.pathsep):
                yield fpath + ext

        fpath, fname = os.path.split(program)
        if fpath:
            if is_exe(program):
                return program
        else:
            for path in os.environ["PATH"].split(os.pathsep):
                exe_file = os.path.join(path, program)
                for candidate in ext_candidates(exe_file):
                    if is_exe(candidate):
                        return candidate
        return None

    def _grab_and_explode(self, urls, name, explode=True):
        """ Special method for grabbing and exploding a package, if necessary.

            Does not verify URLs since these are preverified. Package is
            downloaded to current directory.

            url: list of urls to grab
            name: name of download
            explode: True iff downloaded file should be decompressed

            No return value
        """
        from collections import deque

        self._print_to_screen_and_log(
            "[Configuring] Downloading %s..." % name,
            newline=False,
            carriage_return=True,
        )
        url_deque = deque(urls)
        while url_deque:
            url = url_deque.popleft()
            """Follow redirects (-L), write to file (-O), respect
            content disposition"""
            command = [self.curl_exe, "-L", "-O", url]
            filename = url.rpartition("/")[2]
            print("[Configuring] Downloading {} from {}...".format(name, url))
            try:
                subprocess.check_output(command, stderr=self.log_stream)
            except subprocess.CalledProcessError as e:
                if not url_deque:
                    self._print_to_screen_and_log(
                        (
                            "Error encountered downloading file %s; exit "
                            'code was %d; command invoked was "%s".'
                        )
                        % (url, e.returncode, " ".join(command))
                    )
                    self._print_to_screen_and_log("Make sure web is accessible.")
                    self._bail()
                else:
                    self._print_to_screen_and_log(
                        "[Configuring] Download failed; "
                        "trying alternate URL for %s..." % name,
                        newline=False,
                        carriage_return=True,
                    )
            else:
                if explode:
                    explode_command = None
                    if url[-8:] == ".tar.bz2":
                        explode_command = ["tar", "xvjf", filename]
                    elif url[-7:] == ".tar.gz" or url[-4:] == ".tgz":
                        explode_command = ["tar", "xvzf", filename]
                    elif url[-3:] == ".gz":
                        explode_command = ["gunzip", filename]
                    elif url[-4:] == ".zip":
                        explode_command = ["unzip", filename]
                    if explode_command is not None:
                        self._print_to_screen_and_log(
                            "[Configuring] Extracting %s..." % name,
                            newline=False,
                            carriage_return=True,
                        )
                        try:
                            subprocess.check_output(explode_command)
                        except subprocess.CalledProcessError as e:
                            if not url_deque:
                                self._print_to_screen_and_log(
                                    (
                                        "Error encountered exploding file %s; exit "
                                        'code was %d; command invoked was "%s".'
                                    )
                                    % (
                                        filename,
                                        e.returncode,
                                        " ".join(explode_command),
                                    )
                                )
                                self._bail()
                            else:
                                self._print_to_screen_and_log(
                                    "[Configuring] Extraction failed; "
                                    "trying alternate URL for %s..." % name,
                                    newline=False,
                                    carriage_return=True,
                                )
                                continue
                        finally:
                            try:
                                os.remove(filename)
                            except OSError:
                                break
                break

    def _quote(self, path=None):
        """ Decks path with single quotes if it's not None

            path: path or None

            Return value: None if path is None, else path decked by single
                quotes
        """
        if path is None:
            return "None"
        return "'%s'" % path

    def run(self):
        """ Downloads neoepiscope dependencies. """
        import multiprocessing

        if self.curl_exe is None:
            self.curl_exe = self.check_exe("curl")
            if self.curl_exe is None:
                print_to_screen(
                    "Configuring neoepiscope requires Curl if "
                    "dependencies are to be installed. "
                    "Download it at "
                    "http://curl.haxx.se/download.html and use "
                    "--curl to specify its path, or "
                    "disable installing dependencies with "
                    "--no-dependencies."
                )
                sys.exit(1)
            else:
                self.curl_exe = self.curl_exe.replace(" --help", "")
        self._print_to_screen_and_log("[Configuring] Downloading mhcflurry data...")
        subprocess.call(["mhcflurry-downloads", "fetch"])
        if self.download_dir is None:
            self.download_dir = os.path.join(
                os.path.expanduser("~"), "neoepiscope.data"
            )
            if not self._yes_no_query(
                "Install neoepiscope data in {}?".format(self.download_dir)
            ):
                self.download_dir = os.path.abspath(
                    os.path.expanduser(
                        self._request_path(
                            "Please enter a directory for storing neoepiscope data",
                            use_which=False,
                        )
                    )
                )
        # Download to a temporary directory first, then move to final dest
        temp_install_dir = tempfile.mkdtemp()
        register_cleanup(remove_temporary_directories, [temp_install_dir])
        if os.path.exists(self.download_dir):
            if self._yes_no_query(
                (
                    "The download directory path {dir} already exists.\n    "
                    "* Overwrite {dir}?"
                ).format(dir=self.download_dir)
            ):
                try:
                    shutil.rmtree(self.download_dir)
                except OSError:
                    # Handle this later if directory creation fails
                    pass
                try:
                    os.remove(self.download_dir)
                except OSError:
                    pass
            else:
                print_to_screen("Specify a different download directory")
                sys.exit(0)
        try:
            os.makedirs(self.download_dir)
        except OSError as e:
            self._print_to_screen_and_log(
                (
                    "Problem encountered trying to create "
                    "directory %s for installation. May need "
                    "sudo permissions."
                )
                % self.download_dir
            )
            self._bail()
        else:
            # So it's possible to move temp installation dir there
            os.rmdir(self.download_dir)
            pass
        os.chdir(temp_install_dir)
        if self._yes_no_query("Download Gencode v27 gtf annotation file?"):
            self._grab_and_explode(
                download["Gencode v27 annotation"],
                "Gencode v27 annotation",
                explode=False,
            )
            gencode_v27_temp = os.path.join(temp_install_dir, "gencode_v27")
            gencode_v27 = os.path.join(self.download_dir, "gencode_v27")
            gencode_v27_gtf = os.path.join(
                temp_install_dir,
                os.path.basename(download["Gencode v27 annotation"][0]),
            )
            try:
                os.makedirs(gencode_v27_temp)
            except OSError as e:
                self._print_to_screen_and_log(
                    (
                        "Problem encountered trying to create "
                        "directory %s for installation. May need "
                        "sudo permissions."
                    )
                    % gencode_v27_temp
                )
                self._bail()
            self._print_to_screen_and_log("[Configuring] Indexing Gencode v27...")
            cds_dict = gtf_to_cds(gencode_v27_gtf, gencode_v27_temp)
            cds_to_tree(cds_dict, gencode_v27_temp)
        else:
            gencode_v27 = None
        if self._yes_no_query("Download Gencode v19 gtf annotation file?"):
            self._grab_and_explode(
                download["Gencode v19 annotation"],
                "Gencode v19 annotation",
                explode=False,
            )
            gencode_v19_temp = os.path.join(temp_install_dir, "gencode_v19")
            gencode_v19 = os.path.join(self.download_dir, "gencode_v19")
            gencode_v19_gtf = os.path.join(
                temp_install_dir,
                os.path.basename(download["Gencode v19 annotation"][0]),
            )
            try:
                os.makedirs(gencode_v19_temp)
            except OSError as e:
                self._print_to_screen_and_log(
                    (
                        "Problem encountered trying to create "
                        "directory %s for installation. May need "
                        "sudo permissions."
                    )
                    % gencode_v19_temp
                )
                self._bail()
            self._print_to_screen_and_log("[Configuring] Indexing Gencode v19...")
            cds_dict = gtf_to_cds(gencode_v19_gtf, gencode_v19_temp)
            cds_to_tree(cds_dict, gencode_v19_temp)
        else:
            gencode_v19 = None
        if self._yes_no_query("Download Bowtie NCBI GRCh38 index?"):
            self._grab_and_explode(
                download["Bowtie NCBI GRCh38 index"], "Bowtie NCBI GRCh38 index"
            )
            bowtie_grch38 = os.path.join(
                self.download_dir, "GCA_000001405.15_GRCh38_no_alt_analysis_set"
            )
        else:
            bowtie_grch38 = None
        if self._yes_no_query("Download Bowtie UCSC hg19 index?"):
            self._grab_and_explode(
                download["Bowtie UCSC hg19 index"], "Bowtie UCSC hg19 index"
            )
            bowtie_hg19 = os.path.join(self.download_dir, "hg19")
        else:
            bowtie_hg19 = None
        programs = []
        for program in ["netMHCIIpan v3", "netMHCpan v3", "netMHCpan v4"]:
            if self._yes_no_query(
                (
                    "Do you have an install of {} "
                    "you would like to use for "
                    "binding score predictions with "
                    "neoepiscope?"
                ).format(program)
            ):
                exe = self._request_path(
                    "Please enter the path to your {} executable".format(program),
                    program,
                )
                if is_exe(exe):
                    programs.append(exe)
                else:
                    self._print_to_screen_and_log(
                        (
                            "The path to %s is not an executable. Please "
                            "verify that this is the correct path."
                        )
                        % program
                    )
                    self._bail()
            else:
                programs.append(None)
        if self._yes_no_query(
            "Do you have an install of HapCUT2 "
            "you would like to integrate with neoepiscope?"
        ):
            hapcut_build_dir = self._request_path(
                "Please enter the path to your HapCUT2 build directory", use_which=False
            )
            if is_exe(os.path.join(hapcut_build_dir, "extractHAIRS")) and is_exe(
                os.path.join(hapcut_build_dir, "HAPCUT2")
            ):
                programs.append(os.path.join(hapcut_build_dir, "extractHAIRS"))
                programs.append(os.path.join(hapcut_build_dir, "HAPCUT2"))
            else:
                self._print_to_screen_and_log(
                    "This build directory does not contain the "
                    "relevant HapCUT2 script - please check "
                    "that you have entered the correct path."
                )
                self._bail()
        else:
            for i in range(0, 2):
                programs.append(None)
        # Write paths to exe_paths
        with open(os.path.join(temp_install_dir, "paths.py"), "w") as paths_stream:
            print(
                (
                    """\"""
paths.py
Part of neoepiscope
Defines default paths of neoepiscope's dependencies
None indicates the user didn't install the tool or data
\"""

gencode_v27 = {gencode_v27}
gencode_v19 = {gencode_v19}
bowtie_grch38 = {bowtie_grch38}
bowtie_hg19 = {bowtie_hg19}
netMHCIIpan3 = {netMHCIIpan3}
netMHCpan3 = {netMHCpan3}
netMHCpan4 = {netMHCpan4}
hapcut2_hairs = {hapcut2_hairs}
hapcut2 = {hapcut2}
"""
                ).format(
                    gencode_v27=(
                        "None" if gencode_v27 is None else self._quote(gencode_v27)
                    ),
                    gencode_v19=(
                        "None" if gencode_v19 is None else self._quote(gencode_v19)
                    ),
                    bowtie_grch38=(
                        "None" if bowtie_grch38 is None else self._quote(bowtie_grch38)
                    ),
                    bowtie_hg19=(
                        "None" if bowtie_hg19 is None else self._quote(bowtie_hg19)
                    ),
                    netMHCIIpan3=(
                        "None" if programs[0] is None else self._quote(programs[0])
                    ),
                    netMHCpan3=(
                        "None" if programs[1] is None else self._quote(programs[1])
                    ),
                    netMHCpan4=(
                        "None" if programs[2] is None else self._quote(programs[2])
                    ),
                    hapcut2_hairs=(
                        "None" if programs[3] is None else self._quote(programs[3])
                    ),
                    hapcut2=(
                        "None" if programs[4] is None else self._quote(programs[4])
                    ),
                ),
                file=paths_stream,
            )
        # Move to final directory
        try:
            if os.path.isdir(self.download_dir):
                shutil.rmtree(self.download_dir)
            shutil.copytree(temp_install_dir, self.download_dir)
            shutil.rmtree(temp_install_dir)
            if os.path.isfile(
                os.path.join(
                    sys.prefix,
                    "lib",
                    "".join(
                        [
                            "python",
                            str(sys.version_info.major),
                            ".",
                            str(sys.version_info.minor),
                        ]
                    ),
                    "site-packages",
                    "neoepiscope",
                    "paths.py",
                )
            ):
                os.remove(
                    os.path.join(
                        sys.prefix,
                        "lib",
                        "".join(
                            [
                                "python",
                                str(sys.version_info.major),
                                ".",
                                str(sys.version_info.minor),
                            ]
                        ),
                        "site-packages",
                        "neoepiscope",
                        "paths.py",
                    )
                )
            shutil.copy2(
                os.path.join(self.download_dir, "paths.py"),
                os.path.join(
                    sys.prefix,
                    "lib",
                    "".join(
                        [
                            "python",
                            str(sys.version_info.major),
                            ".",
                            str(sys.version_info.minor),
                        ]
                    ),
                    "site-packages",
                    "neoepiscope",
                ),
            )
            os.remove(os.path.join(self.download_dir, "paths.py"))
        except Exception as e:
            self._print_to_screen_and_log(
                (
                    'Problem "%s" encountered moving '
                    "temporary installation directory "
                    "%s to final destination %s."
                )
                % (e, temp_install_dir, self.download_dir)
            )
            self._bail()
        self._print_to_screen_and_log("Configured neoepiscope.")
        self.finished = True



#!/usr/bin/env python
# coding=utf-8
"""
bowtie_index.py

Includes class for grabbing genome sequence from Bowtie index.

The MIT License (MIT)
Taken from Rail-RNA, which is copyright (c) 2015 
                    Abhinav Nellore, Leonardo Collado-Torres,
                    Andrew Jaffe, James Morton, Jacob Pritt,
                    José Alquicira-Hernández,
                    Christopher Wilks,
                    Jeffrey T. Leek, and Ben Langmead.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import absolute_import, division, print_function
import os
import struct
import mmap
from operator import itemgetter
from collections import defaultdict
from bisect import bisect_right
import sys

# Python 2-3 compatibility
try:
    xrange
except NameError:
    xrange = range

if sys.version_info[0] > 2:

    def byte_readline(fh):
        return fh.readline().decode(encoding="UTF-8")

    # we need a version of ord that does nothing in python 3
    # but regular ord is still needed on refname reads thanks to
    # our byte_readline
    def ord2or3(chr):
        return chr


else:

    def byte_readline(fh):
        return fh.readline()

    def ord2or3(chr):
        return ord(chr)


class BowtieIndexReference(object):
    """
    Given prefix of a Bowtie index, parses the reference names, parses the
    extents of the unambiguous stretches, and memory-maps the file containing
    the unambiguous-stretch sequences.  get_stretch member function can
    retrieve stretches of characters from the reference, even if the stretch
    contains ambiguous characters.
    """

    def __init__(self, idx_prefix):

        # Open file handles
        if os.path.exists(idx_prefix + ".3.ebwt"):
            # Small index (32-bit offsets)
            fh1 = open(idx_prefix + ".1.ebwt", "rb")  # for ref names
            fh3 = open(idx_prefix + ".3.ebwt", "rb")  # for stretch extents
            fh4 = open(idx_prefix + ".4.ebwt", "rb")  # for unambiguous sequence
            sz, struct_unsigned = 4, struct.Struct("I")
        else:
            raise RuntimeError('No Bowtie index files with prefix "%s"' % idx_prefix)

        #
        # Parse .1.bt2 file
        #
        one = struct.unpack("<i", fh1.read(4))[0]
        assert one == 1

        ln = struct_unsigned.unpack(fh1.read(sz))[0]
        line_rate = struct.unpack("<i", fh1.read(4))[0]
        lines_per_side = struct.unpack("<i", fh1.read(4))[0]
        _ = struct.unpack("<i", fh1.read(4))[0]
        ftab_chars = struct.unpack("<i", fh1.read(4))[0]
        _ = struct.unpack("<i", fh1.read(4))[0]

        nref = struct_unsigned.unpack(fh1.read(sz))[0]
        # get ref lengths
        reference_length_list = []
        for i in xrange(nref):
            reference_length_list.append(struct.unpack("<i", fh1.read(sz))[0])

        nfrag = struct_unsigned.unpack(fh1.read(sz))[0]
        # skip rstarts
        fh1.seek(nfrag * sz * 3, 1)

        # skip ebwt
        bwt_sz = ln // 4 + 1
        line_sz = 1 << line_rate
        side_sz = line_sz * lines_per_side
        side_bwt_sz = side_sz - 8
        num_side_pairs = (bwt_sz + (2 * side_bwt_sz) - 1) // (2 * side_bwt_sz)
        ebwt_tot_len = num_side_pairs * 2 * side_sz
        fh1.seek(ebwt_tot_len, 1)

        # skip zOff
        fh1.seek(sz, 1)

        # skip fchr
        fh1.seek(5 * sz, 1)

        # skip ftab
        ftab_len = (1 << (ftab_chars * 2)) + 1
        fh1.seek(ftab_len * sz, 1)

        # skip eftab
        eftab_len = ftab_chars * 2
        fh1.seek(eftab_len * sz, 1)

        refnames = []
        while True:
            refname = byte_readline(fh1)
            if len(refname) == 0 or ord(refname[0]) == 0:
                break
            refnames.append(refname.split()[0])
        assert len(refnames) == nref

        #
        # Parse .3.bt2 file
        #
        one = struct.unpack("<i", fh3.read(4))[0]
        assert one == 1

        nrecs = struct_unsigned.unpack(fh3.read(sz))[0]

        running_unambig, running_length = 0, 0
        self.recs = defaultdict(list)
        self.offset_in_ref = defaultdict(list)
        self.unambig_preceding = defaultdict(list)
        length = {}

        ref_id, ref_namenrecs_added = 0, None
        for i in xrange(nrecs):
            off = struct_unsigned.unpack(fh3.read(sz))[0]
            ln = struct_unsigned.unpack(fh3.read(sz))[0]
            first_of_chromosome = ord(fh3.read(1)) != 0
            if first_of_chromosome:
                if i > 0:
                    length[ref_name] = running_length
                ref_name = refnames[ref_id]
                ref_id += 1
                running_length = 0
            assert ref_name is not None
            self.recs[ref_name].append((off, ln, first_of_chromosome))
            self.offset_in_ref[ref_name].append(running_length)
            self.unambig_preceding[ref_name].append(running_unambig)
            running_length += off + ln
            running_unambig += ln

        length[ref_name] = running_length
        assert nrecs == sum(map(len, self.recs.values()))

        #
        # Memory-map the .4.bt2 file
        #
        ln_bytes = (running_unambig + 3) // 4
        self.fh4mm = mmap.mmap(
            fh4.fileno(), ln_bytes, flags=mmap.MAP_SHARED, prot=mmap.PROT_READ
        )

        # These are per-reference
        self.length = length
        self.refnames = refnames

        # To facilitate sorting reference names in order of descending length
        sorted_rnames = sorted(
            self.length.items(), key=lambda x: itemgetter(1)(x), reverse=True
        )
        """A case-sensitive sort is also necessary here because new versions of
        bedGraphToBigWig complain on encountering a nonlexicographic sort
        order."""
        lexicographically_sorted_rnames = sorted(
            self.length.items(), key=lambda x: itemgetter(0)(x)
        )
        self.rname_to_string, self.l_rname_to_string = {}, {}
        self.string_to_rname, self.l_string_to_rname = {}, {}
        for i, (rname, _) in enumerate(sorted_rnames):
            rname_string = "%012d" % i
            self.rname_to_string[rname] = rname_string
            self.string_to_rname[rname_string] = rname
        for i, (rname, _) in enumerate(lexicographically_sorted_rnames):
            rname_string = "%012d" % i
            self.l_rname_to_string[rname] = rname_string
            self.l_string_to_rname[rname_string] = rname
        # Handle unmapped reads
        unmapped_string = "%012d" % len(sorted_rnames)
        self.rname_to_string["*"] = unmapped_string
        self.string_to_rname[unmapped_string] = "*"

        # For compatibility
        self.rname_lengths = self.length
        fh1.close()
        fh3.close()
        fh4.close()

    def get_stretch(self, ref_id, ref_off, count):
        """
        Return a stretch of characters from the reference, retrieved
        from the Bowtie index.

        @param ref_id: name of ref seq, up to & excluding whitespace
        @param ref_off: offset into reference, 0-based
        @param count: # of characters
        @return: string extracted from reference
        """
        assert ref_id in self.recs
        # Account for negative reference offsets by padding with Ns
        N_count = min(abs(min(ref_off, 0)), count)
        stretch = ["N"] * N_count
        count -= N_count
        if not count:
            return "".join(stretch)
        ref_off = max(ref_off, 0)
        starting_rec = bisect_right(self.offset_in_ref[ref_id], ref_off) - 1
        assert starting_rec >= 0
        off = self.offset_in_ref[ref_id][starting_rec]
        buf_off = self.unambig_preceding[ref_id][starting_rec]
        # Naive to scan these records linearly; obvious speedup is binary search
        for rec in self.recs[ref_id][starting_rec:]:
            off += rec[0]
            while ref_off < off and count > 0:
                stretch.append("N")
                count -= 1
                ref_off += 1
            if count == 0:
                break
            if ref_off < off + rec[1]:
                # stretch extends through part of the unambiguous stretch
                buf_off += ref_off - off
            else:
                buf_off += rec[1]
            off += rec[1]
            while ref_off < off and count > 0:
                buf_elt = buf_off >> 2
                shift_amt = (buf_off & 3) << 1
                stretch.append("ACGT"[(ord2or3(self.fh4mm[buf_elt]) >> shift_amt) & 3])
                buf_off += 1
                count -= 1
                ref_off += 1
            if count == 0:
                break
        # If the requested stretch went past the last unambiguous
        # character in the chromosome, pad with Ns
        while count > 0:
            count -= 1
            stretch.append("N")
        return "".join(stretch)


def which(program):
    def is_exe(fp):
        return os.path.isfile(fp) and os.access(fp, os.X_OK)

    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file

    return None


if __name__ == "__main__":

    import sys
    import unittest
    import argparse
    from tempfile import mkdtemp
    from shutil import rmtree

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--test", action="store_const", const=True, default=False, help="Do unit tests"
    )

    args = parser.parse_args()

    if args.test:
        import unittest

        class TestBowtieIndexReference(unittest.TestCase):
            def setUp(self):
                self.tmpdir = mkdtemp()
                self.fa_fn_1 = os.path.join(self.tmpdir, "tmp1.fa")
                with open(self.fa_fn_1, "w") as fh:
                    fh.write(
                        """>short_name1 with some stuff after whitespace
ACGTACGTACGTACGTACGTACGTACGTACGTACGTACGT
ACGTACGTACGTACGTACGTACGTACGTACGTACGTACGT
A
>short_name4
NNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNN
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
NNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNN
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
NNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNN
GGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGG
NNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNN
TT
>short_name2 with some stuff after whitespace
CAGTCAGTCAGTCAGTCAGTCAGTCAGTCAGTCAGTCAGT
CAGTCAGTCAGTCAGTCAGTCAGTCAGTCAGTCAGTCAGT
>short_name3 with some stuff after whitespace
CA
"""
                    )
                assert which("bowtie-build") is not None
                os.system(
                    "bowtie-build %s %s >/dev/null" % (self.fa_fn_1, self.fa_fn_1)
                )

            def tearDown(self):
                rmtree(self.tmpdir)

            def test1(self):
                ref = BowtieIndexReference(self.fa_fn_1)
                self.assertEqual("ACGTACGTAC", ref.get_stretch("short_name1", 0, 10))
                self.assertEqual("ACGTACGTAC", ref.get_stretch("short_name1", 40, 10))
                self.assertEqual("ANNNNNNNNN", ref.get_stretch("short_name1", 80, 10))

                self.assertEqual("CAGTCAGTCA", ref.get_stretch("short_name2", 0, 10))
                self.assertEqual("CAGTCAGTCA", ref.get_stretch("short_name2", 40, 10))
                self.assertEqual("NNNNNNNNNN", ref.get_stretch("short_name2", 80, 10))

                self.assertEqual("CANNNNNNNN", ref.get_stretch("short_name3", 0, 10))

            def test2(self):
                ref = BowtieIndexReference(self.fa_fn_1)
                self.assertEqual("CAGTCAGTCA", ref.get_stretch("short_name2", 0, 10))
                self.assertEqual("AGTCAGTCAGT", ref.get_stretch("short_name2", 1, 11))
                self.assertEqual("GTCAGTCAGTCA", ref.get_stretch("short_name2", 2, 12))
                self.assertEqual("TCAGTCAGTCAGT", ref.get_stretch("short_name2", 3, 13))

                self.assertEqual("TACGTACGTA", ref.get_stretch("short_name1", 71, 10))
                self.assertEqual("ACGTACGTANN", ref.get_stretch("short_name1", 72, 11))
                self.assertEqual("CGTACGTANNNN", ref.get_stretch("short_name1", 73, 12))

            def test3(self):
                ref = BowtieIndexReference(self.fa_fn_1)
                self.assertEqual(
                    "NNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNN",
                    ref.get_stretch("short_name4", 0, 40),
                )
                self.assertEqual(
                    "NNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNA",
                    ref.get_stretch("short_name4", 1, 40),
                )
                self.assertEqual("AAAA", ref.get_stretch("short_name4", 41, 4))
                self.assertEqual(
                    "NNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNTT",
                    ref.get_stretch("short_name4", 240, 42),
                )

            def test4(self):
                ref = BowtieIndexReference(self.fa_fn_1)
                # Test that all refname lengths are accurate
                self.assertEqual(ref.length["short_name1"], 81)
                self.assertEqual(ref.length["short_name4"], 282)
                self.assertEqual(ref.length["short_name2"], 80)
                self.assertEqual(ref.length["short_name3"], 2)

            def test_off_reference_values(self):
                ref = BowtieIndexReference(self.fa_fn_1)
                self.assertEqual("NNNACG", ref.get_stretch("short_name1", -3, 6))
                self.assertEqual("NNNNN", ref.get_stretch("short_name1", -20, 5))
                self.assertEqual("NNNNNNNNN", ref.get_stretch("short_name1", 85, 9))
                self.assertEqual("ANNNNNNNN", ref.get_stretch("short_name1", 80, 9))

        unittest.main(argv=[sys.argv[0]])



#!/usr/bin/env python
# coding=utf-8
"""
binding_scores.py

Part of neoepiscope
Includes functions for grabbing MHC binding scores from third-party software.

The MIT License (MIT)
Copyright (c) 2018 Mary A. Wood, Austin Nguyen,
                   Abhinav Nellore, and Reid Thompson

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import absolute_import, division, print_function
import os
import warnings
import tempfile
import pickle
import subprocess

neoepiscope_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))


def get_affinity_netMHCIIpan(
    peptides, allele, netmhciipan, version, scores, remove_files=True
):
    """ Obtains binding affinities from list of peptides

        peptides: peptides of interest (list of strings)
        allele: Allele to use for binding affinity (string)
        netmhciipan: path to netMHCIIpan executable
        version: version of netMHCIIpan
        scores: list of scoring methods
        remove_files: option to remove intermediate files

        Return value: affinities (a list of binding affinities
                        as strings)
    """
    files_to_remove = []
    try:
        # Check that allele is valid for method
        with open(
            os.path.join(
                os.path.join(neoepiscope_dir, "neoepiscope", "availableAlleles.pickle")
            ),
            "rb",
        ) as allele_stream:
            avail_alleles = pickle.load(allele_stream)
        # Homogenize format
        if "DRB" in allele:
            allele = allele.replace("HLA-", "").replace(":", "").replace("*", "_")
        elif "DP" in allele or "DQ" in allele:
            allele = allele.replace(":", "").replace("*", "")
        if allele not in avail_alleles["".join(["netMHCIIpan", str(version)])]:
            warnings.warn(
                " ".join([allele, "is not a valid allele for netMHCIIpan"]), Warning
            )
            if len(scores) == 1:
                return [(peptides[i], "NA") for i in range(0, len(peptides))]
            else:
                return [(peptides[i], "NA", "NA") for i in range(0, len(peptides))]
        allele = allele.replace("*", "_").replace(":", "")
        # Establish return list and sample id
        sample_id = ".".join(
            [peptides[0], str(len(peptides)), allele, "netmhciipan", version]
        )
        affinities = []
        # Write one peptide per line to a temporary file for
        #   input if peptide length is at least 9
        # Count instances of smaller peptides
        na_count = 0
        peptide_file = tempfile.mkstemp(
            suffix=".peptides", prefix="".join([sample_id, "."]), text=True
        )[1]
        files_to_remove.append(peptide_file)
        with open(peptide_file, "w") as f:
            for sequence in peptides:
                if len(sequence) >= 9:
                    print(sequence, file=f)
                else:
                    na_count += 1
        if na_count > 0:
            warnings.warn(
                " ".join(
                    [
                        str(na_count),
                        "peptides not compatible with",
                        "netMHCIIpan will not receive score",
                    ]
                ),
                Warning,
            )
        # Establish temporary file to hold output
        mhc_out = tempfile.mkstemp(
            suffix=".netMHCIIpan.out", prefix="".join([sample_id, "."]), text=True
        )[1]
        files_to_remove.append(mhc_out)
        # Run netMHCIIpan
        subprocess.check_call(
            [
                netmhciipan,
                "-a",
                allele,
                "-inptype",
                "1",
                "-xls",
                "-xlsfile",
                mhc_out,
                "-f",
                peptide_file,
            ]
        )
        # Retrieve scores for valid peptides
        score_dict = {}
        with open(mhc_out, "r") as f:
            # Skip headers
            f.readline()
            f.readline()
            for line in f:
                # token 1 is peptide; token 4 is affinity; token[5] is rank
                tokens = line.strip("\n").split("\t")
                if sorted(scores) == ["affinity", "rank"]:
                    score_dict[tokens[1]] = (tokens[4], tokens[5])
                elif sorted(scores) == ["affinity"]:
                    score_dict[tokens[1]] = (tokens[4],)
                elif sorted(scores) == ["rank"]:
                    score_dict[tokens[1]] = (tokens[5],)
        # Produce list of scores for valid peptides
        # Invalid peptides receive "NA" score
        for sequence in peptides:
            if sequence in score_dict:
                nM = (sequence,) + score_dict[sequence]
            else:
                if len(scores) == 1:
                    nM = (sequence, "NA")
                else:
                    nM = (sequence, "NA", "NA")
            affinities.append(nM)
        return affinities
    finally:
        if remove_files:
            for file_to_remove in files_to_remove:
                os.remove(file_to_remove)


def get_affinity_mhcflurry(peptides, allele, scores, version, remove_files=True):
    """ Obtains binding affinities from list of peptides

        peptides: peptides of interest (list of strings)
        allele: Allele to use for binding affinity (string)
        scores: list of scoring methods
        version: version of mhc-flurry
        remove_files: option to remove intermediate files

        Return value: affinities (a list of binding affinities
                        as strings)
    """
    files_to_remove = []
    try:
        # Check that allele is valid for method
        with open(
            os.path.join(neoepiscope_dir, "neoepiscope", "availableAlleles.pickle"),
            "rb",
        ) as allele_stream:
            avail_alleles = pickle.load(allele_stream)
        if allele not in avail_alleles["mhcflurry"]:
            warnings.warn(
                " ".join([allele, "is not a valid allele for mhcflurry"]), Warning
            )
            score_form = tuple(["NA" for i in range(0, len(scores))])
            return [(peptides[i],) + score_form for i in range(0, len(peptides))]
        # Establish return list and sample id
        sample_id = ".".join(
            [peptides[0], str(len(peptides)), allele, "mhcflurry", version]
        )
        affinities = []
        # Write one peptide per line to a temporary file for
        #   input if peptide length is at least 9
        # Count instances of smaller peptides
        # Establish temporary file to hold output
        peptide_file = tempfile.mkstemp(
            suffix=".csv", prefix="".join([sample_id, "."]), text=True
        )[1]
        na_count = 0
        with open(peptide_file, "w") as f:
            f.write("allele,peptide\n")
            for sequence in peptides:
                if len(sequence) < 8 or len(sequence) > 15:
                    na_count += 1
                else:
                    print("".join([allele, ",", sequence, "\n"]), file=f)
        if na_count > 0:
            warnings.warn(
                " ".join(
                    [
                        str(na_count),
                        "peptides not compatible with",
                        "mhcflurry will not receive score",
                    ]
                ),
                Warning,
            )
        # Establish temporary file to hold output
        mhc_out = tempfile.mkstemp(
            suffix=".mhcflurry.out", prefix="".join([sample_id, "."]), text=True
        )[1]
        files_to_remove.append(mhc_out)
        # Run netMHCIIpan
        command = ["mhcflurry-predict", "--out", mhc_out, peptide_file]
        subprocess.check_call(command)
        # Retrieve scores for valid peptides
        score_dict = {}
        with open(mhc_out, "r") as f:
            # Skip headers
            f.readline()
            for line in f:
                # token 1 is peptide; token 4 is affinity; token[5] is rank
                tokens = line.strip("\n").split(",")
                if sorted(scores) == ["affinity", "rank"]:
                    score_dict[tokens[1]] = (tokens[2], tokens[5])
                elif sorted(scores) == ["affinity", "high", "low", "rank"]:
                    score_dict[tokens[1]] = (tokens[2], tokens[4], tokens[3], tokens[5])
                elif sorted(scores) == ["affinity", "high", "low"]:
                    score_dict[tokens[1]] = (tokens[2], tokens[4], tokens[3])
                elif sorted(scores) == ["affinity", "high", "rank"]:
                    score_dict[tokens[1]] = (tokens[2], tokens[4], tokens[5])
                elif sorted(scores) == ["affinity", "low", "rank"]:
                    score_dict[tokens[1]] = (tokens[2], tokens[3], tokens[5])
                elif sorted(scores) == ["high", "low", "rank"]:
                    score_dict[tokens[1]] = (tokens[4], tokens[3], tokens[3])
                elif sorted(scores) == ["affinity", "high"]:
                    score_dict[tokens[1]] = (tokens[2], tokens[4])
                elif sorted(scores) == ["affinity", "low"]:
                    score_dict[tokens[1]] = (tokens[2], tokens[3])
                elif sorted(scores) == ["high", "low"]:
                    score_dict[tokens[1]] = (tokens[4], tokens[3])
                elif sorted(scores) == ["high", "rank"]:
                    score_dict[tokens[1]] = (tokens[4], tokens[5])
                elif sorted(scores) == ["low", "rank"]:
                    score_dict[tokens[1]] = (tokens[3], tokens[5])
                elif sorted(scores) == ["affinity"]:
                    score_dict[tokens[1]] = (tokens[2],)
                elif sorted(scores) == ["rank"]:
                    score_dict[tokens[1]] = (tokens[5],)
                elif sorted(scores) == ["high"]:
                    score_dict[tokens[1]] = (tokens[4],)
                elif sorted(scores) == ["low"]:
                    score_dict[tokens[1]] = (tokens[3],)
        # Produce list of scores for valid peptides
        # Invalid peptides receive "NA" score
        for sequence in peptides:
            if sequence in score_dict:
                nM = (sequence,) + score_dict[sequence]
            else:
                nM = (sequence,) + tuple(["NA" for i in range(0, len(scores))])
            affinities.append(nM)
        return affinities
    finally:
        if remove_files:
            for file_to_remove in files_to_remove:
                os.remove(file_to_remove)


def get_affinity_netMHCpan(
    peptides, allele, netmhcpan, version, scores, remove_files=True
):
    """ Obtains binding affinities from list of peptides

        peptides: peptides of interest (list of strings)
        allele: allele to use for binding affinity
                    (string, format HLA-A02:01)
        netmhcpan: path to netMHCpan executable
        version: version of netMHCpan software
        scores: list of scoring methods
        remove_files: option to remove intermediate files

        Return value: affinities (a list of binding affinities
                        as strings)
    """
    files_to_remove = []
    try:
        # Check that allele is valid for method
        with open(
            os.path.join(neoepiscope_dir, "neoepiscope", "availableAlleles.pickle"),
            "rb",
        ) as allele_stream:
            avail_alleles = pickle.load(allele_stream)
        allele = allele.replace("*", "")
        if allele not in avail_alleles["".join(["netMHCpan", str(version)])]:
            warnings.warn(
                " ".join([allele, "is not a valid allele for netMHCpan"]), Warning
            )
            if len(scores) == 1:
                return [(peptides[i], "NA") for i in range(0, len(peptides))]
            else:
                return [(peptides[i], "NA", "NA") for i in range(0, len(peptides))]
        # Establish return list and sample id
        sample_id = ".".join(
            [peptides[0], str(len(peptides)), allele, "netmhcpan", version]
        )
        affinities = []
        # Write one peptide per line to a temporary file for input
        peptide_file = tempfile.mkstemp(
            suffix=".peptides", prefix="".join([sample_id, "."]), text=True
        )[1]
        files_to_remove.append(peptide_file)
        with open(peptide_file, "w") as f:
            for sequence in peptides:
                print(sequence, file=f)
        # Establish temporary file to hold output
        mhc_out = tempfile.mkstemp(
            suffix=".netMHCpan.out", prefix="".join([sample_id, "."]), text=True
        )[1]
        files_to_remove.append(mhc_out)
        err_file = tempfile.mkstemp(
            suffix=".netMHCpan.err", prefix="".join([sample_id, "."]), text=True
        )[1]
        files_to_remove.append(err_file)
        with open(err_file, "w") as e:
            # Run netMHCpan
            if version == "3":
                subprocess.check_call(
                    [
                        netmhcpan,
                        "-a",
                        allele,
                        "-inptype",
                        "1",
                        "-p",
                        "-xls",
                        "-xlsfile",
                        mhc_out,
                        peptide_file,
                    ],
                    stderr=e,
                )
            elif version == "4":
                subprocess.check_call(
                    [
                        netmhcpan,
                        "-BA",
                        "-a",
                        allele,
                        "-inptype",
                        "1",
                        "-p",
                        "-xls",
                        "-xlsfile",
                        mhc_out,
                        peptide_file,
                    ],
                    stderr=e,
                )
        with open(mhc_out, "r") as f:
            f.readline()
            f.readline()
            for i in range(0, len(peptides)):
                tokens = f.readline().strip("\n").split("\t")
                # for v3, tokens[5] is affinity, tokens[6] is rank
                # for v4, tokens[6] is affinty, tokens[7] is rank
                if sorted(scores) == ["affinity", "rank"]:
                    if version == "3":
                        nM = (peptides[i], tokens[5], tokens[6])
                    elif version == "4":
                        nM = (peptides[i], tokens[6], tokens[7])
                elif sorted(scores) == ["affinity"]:
                    if version == "3":
                        nM = (peptides[i], tokens[5])
                    elif version == "4":
                        nM = (peptides[i], tokens[6])
                elif sorted(scores) == ["rank"]:
                    if version == "3":
                        nM = (peptides[i], tokens[6])
                    elif version == "4":
                        nM = (peptides[i], tokens[7])
                affinities.append(nM)
        return affinities
    finally:
        # Remove temporary files
        if remove_files:
            for file_to_remove in files_to_remove:
                os.remove(file_to_remove)


def get_affinity_mhcnuggets(peptides, allele, version, remove_files=True):
    """ Obtains binding affinities from list of peptides

        peptides: peptides of interest (list of strings)
        allele: Allele to use for binding affinity (string)
        scores: list of scoring methods
        version: version of mhcnuggets
        remove_files: option to remove intermediate files

        Return value: affinities (a list of binding affinities
                        as strings)
    """
    from mhcnuggets.src.predict import predict

    files_to_remove = []
    try:
        # Check that allele is valid for method
        with open(
            os.path.join(neoepiscope_dir, "neoepiscope", "availableAlleles.pickle"),
            "rb",
        ) as allele_stream:
            avail_alleles = pickle.load(allele_stream)
        # Check that allele is valid for method
        allele = allele.replace("*", "")
        if allele in avail_alleles["mhcnuggets_mhcI"]:
            allele_class = "I"
            max_length = 15
        elif allele in avail_alleles["mhcnuggets_mhcII"]:
            allele_class = "II"
            max_length = 30
        else:
            warnings.warn(
                " ".join([allele, "is not a valid allele for mhcnuggets"]), Warning
            )
            return [(peptides[i], "NA") for i in range(0, len(peptides))]
        # Establish return list and sample id
        sample_id = ".".join(
            [peptides[0], str(len(peptides)), allele, "mhcnuggets", version]
        )
        affinities = []
        # Write one peptide per line to a temporary file for
        #   input if peptide length is at least 9
        # Count instances of smaller peptides
        # Establish temporary file to hold output
        peptide_file = tempfile.mkstemp(
            suffix=".txt", prefix="".join([sample_id, "."]), text=True
        )[1]
        files_to_remove.append(peptide_file)
        na_count = 0
        with open(peptide_file, "w") as f:
            for sequence in peptides:
                if len(sequence) > max_length:
                    na_count += 1
                else:
                    print(sequence, file=f)
        if na_count > 0:
            warnings.warn(
                " ".join(
                    [
                        str(na_count),
                        "peptides not compatible with",
                        "mhcnuggets will not receive score",
                    ]
                ),
                Warning,
            )
        # Establish temporary file to hold output
        mhc_out = tempfile.mkstemp(
            suffix=".mhcnuggets.out", prefix="".join([sample_id, "."]), text=True
        )[1]
        files_to_remove.append(mhc_out)
        # Run mhcnuggets
        predict(
            class_=allele_class, peptides_path=peptide_file, mhc=allele, output=mhc_out
        )
        # Retrieve scores for valid peptides
        score_dict = {}
        with open(mhc_out, "r") as f:
            # Skip headers
            f.readline()
            for line in f:
                tokens = line.strip("\n").split(",")
                score_dict[tokens[0]] = tokens[1]
        # Produce list of scores for valid peptides
        # Invalid peptides receive "NA" score
        for sequence in peptides:
            if sequence in score_dict:
                nM = (sequence, score_dict[sequence])
            else:
                nM = (sequence, "NA")
            affinities.append(nM)
        return affinities
    finally:
        if remove_files:
            for file_to_remove in files_to_remove:
                os.remove(file_to_remove)


def gather_binding_scores(neoepitopes, tool_dict, hla_alleles):
    """ Adds binding scores from desired programs to neoepitope metadata

        neoepitopes: dictionary linking neoepitopes to their metadata
        tool_dict: dictionary storing prediction tool data
        hla_alleles: list of HLA alleles used for binding predictions

        Return value: dictionary linking neoepitopes to their metadata,
            which now includes binding scores
    """
    for allele in hla_alleles:
        for tool in sorted(tool_dict.keys()):
            if tool == "mhcflurry1":
                binding_scores = get_affinity_mhcflurry(
                    list(neoepitopes.keys()),
                    allele,
                    tool_dict[tool][1],
                    "1",
                    remove_files=True,
                )
            elif tool == "mhcnuggets2":
                binding_scores = get_affinity_mhcnuggets(
                    list(neoepitopes.keys()), allele, "2", remove_files=True
                )
            elif tool == "netMHCIIpan3":
                binding_scores = get_affinity_netMHCIIpan(
                    list(neoepitopes.keys()),
                    allele,
                    tool_dict[tool][0],
                    "3",
                    tool_dict[tool][1],
                    remove_files=True,
                )
            elif tool == "netMHCpan3":
                binding_scores = get_affinity_netMHCpan(
                    list(neoepitopes.keys()),
                    allele,
                    tool_dict[tool][0],
                    "3",
                    tool_dict[tool][1],
                    remove_files=True,
                )
            elif tool == "netMHCpan4":
                binding_scores = get_affinity_netMHCpan(
                    list(neoepitopes.keys()),
                    allele,
                    tool_dict[tool][0],
                    "4",
                    tool_dict[tool][1],
                    remove_files=True,
                )
            for score in binding_scores:
                meta_data = neoepitopes[score[0]]
                for i in range(0, len(meta_data)):
                    neoepitopes[score[0]][i] = meta_data[i] + score[1:]
    return neoepitopes



#!/usr/bin/env python
# coding=utf-8
"""
neoepiscope

Identifies neoepitopes from DNA-seq, VCF, GTF, and Bowtie index.

The MIT License (MIT)
Copyright (c) 2018 Mary A. Wood, Austin Nguyen,
                   Abhinav Nellore, and Reid F. Thompson

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import absolute_import, division, print_function
import argparse
from . import bowtie_index
import sys
import string
import copy
import pickle
import copy
import random
import re
import os
import collections
import tempfile
import subprocess
import warnings
from . import paths
from .transcript import (
    Transcript,
    gtf_to_cds,
    cds_to_tree,
    get_transcripts_from_tree,
    process_haplotypes,
    get_peptides_from_transcripts,
)
from .binding_scores import gather_binding_scores
from .file_processing import (
    adjust_tumor_column,
    combine_vcf,
    prep_hapcut_output,
    which,
    get_vaf_pos,
    write_results,
)
from operator import itemgetter
from intervaltree import Interval, IntervalTree

_help_intro = (
    """neoepiscope searches for neoepitopes using tumor/normal DNA-seq data."""
)


def help_formatter(prog):
    """ So formatter_class's max_help_position can be changed. """
    return argparse.HelpFormatter(prog, max_help_position=40)


def main():
    """ Entry point for neoepiscope software """
    parser = argparse.ArgumentParser(
        description=_help_intro, formatter_class=help_formatter
    )
    subparsers = parser.add_subparsers(
        help=(
            'subcommands; add "-h" or "--help" ' "after a subcommand for its parameters"
        ),
        dest="subparser_name",
    )
    index_parser = subparsers.add_parser(
        "index",
        help=(
            "produces pickled dictionaries "
            "linking transcripts to intervals and "
            " CDS lines in a GTF"
        ),
    )
    swap_parser = subparsers.add_parser(
        "swap",
        help=(
            "swaps tumor and normal columns "
            "in a somatic vcf if necessary for "
            "proper HapCUT2 results"
        ),
    )
    merge_parser = subparsers.add_parser(
        "merge",
        help=(
            "merges germline and somatic "
            "VCFS for combined mutation "
            "phasing with HAPCUT2"
        ),
    )
    download_parser = subparsers.add_parser("download", help="downloads dependencies")
    prep_parser = subparsers.add_parser(
        "prep", help=("combines HAPCUT2 output with unphased variants for call mode")
    )
    call_parser = subparsers.add_parser("call", help="calls neoepitopes")
    # Index parser options (produces pickled dictionaries for transcript data)
    index_parser.add_argument(
        "-g", "--gtf", type=str, required=True, help="input path to GTF file"
    )
    index_parser.add_argument(
        "-d",
        "--dicts",
        type=str,
        required=True,
        help="output path to pickled CDS dictionary directory",
    )
    # Swap parser options (swaps columns in somatic VCF)
    swap_parser.add_argument(
        "-i", "--input", type=str, required=True, help="input path to somatic VCF"
    )
    swap_parser.add_argument(
        "-o",
        "--output",
        type=str,
        required=False,
        default="-",
        help="output path to column-swapped VCF; use - for stdout",
    )
    # Merger parser options (merges somatic and germline VCFs)
    merge_parser.add_argument(
        "-g", "--germline", type=str, required=True, help="input path to germline VCF"
    )
    merge_parser.add_argument(
        "-s", "--somatic", type=str, required=True, help="input path to somatic VCF"
    )
    merge_parser.add_argument(
        "-o",
        "--output",
        type=str,
        required=False,
        default="-",
        help="output path to combined VCF; use - for stdout",
    )
    # Prep parser options (adds unphased mutations as their own haplotype)
    prep_parser.add_argument("-v", "--vcf", type=str, required=True, help="input VCF")
    prep_parser.add_argument(
        "-c",
        "--hapcut2-output",
        type=str,
        required=False,
        help="path to output file of HAPCUT2 run on input VCF",
    )
    prep_parser.add_argument(
        "-o",
        "--output",
        type=str,
        required=False,
        default="-",
        help="path to output file to be input to call mode; use - for stdout",
    )
    # Call parser options (calls neoepitopes)
    call_parser.add_argument(
        "-x",
        "--bowtie-index",
        type=str,
        required=False,
        help="path to Bowtie index basename",
    )
    call_parser.add_argument(
        "-v", "--vcf", type=str, required=False, help="input path to VCF"
    )
    call_parser.add_argument(
        "-d",
        "--dicts",
        type=str,
        required=False,
        help="input path to pickled CDS dictionary directory",
    )
    call_parser.add_argument(
        "-c",
        "--merged-hapcut2-output",
        type=str,
        required=False,
        default="-",
        help="path to output of prep subcommand; use - for stdin",
    )
    call_parser.add_argument(
        "-k",
        "--kmer-size",
        type=str,
        required=False,
        default="8,11",
        help="kmer size for epitope calculation",
    )
    call_parser.add_argument(
        "-p",
        "--affinity-predictor",
        type=str,
        nargs=3,
        required=False,
        action="append",
        default=[["mhcflurry", "1", "affinity,rank"]],
        help="binding affinity prediction software,"
        "associated version number, and scoring method(s) "
        "(e.g. -p netMHCpan 4 rank,affinity); "
        "for multiple programs, repeat the argument; "
        "see documentation for details",
    )
    call_parser.add_argument(
        "-n",
        "--no-affinity",
        required=False,
        action="store_true",
        help="do not run binding affinity predictions; overrides any "
        "binding affinity prediction tools specified via "
        "--affinity-predictor option",
    )
    call_parser.add_argument(
        "-a",
        "--alleles",
        type=str,
        required=False,
        help="comma separated list of alleles; "
        "see documentation online for more information",
    )
    call_parser.add_argument(
        "-o",
        "--output",
        type=str,
        required=False,
        default="-",
        help="path to output file; use - for stdout",
    )
    call_parser.add_argument(
        "-f",
        "--fasta",
        required=False,
        action="store_true",
        help="produce additional fasta output; see documentation",
    )
    call_parser.add_argument(
        "-u",
        "--upstream-atgs",
        type=str,
        required=False,
        default="none",
        help="how to handle upstream start codons, see "
        "documentation online for more information",
    )
    call_parser.add_argument(
        "-g",
        "--germline",
        type=str,
        required=False,
        default="background",
        help="how to handle germline mutations in "
        "neoepitope enumeration; documentation online for more information",
    )
    call_parser.add_argument(
        "-s",
        "--somatic",
        type=str,
        required=False,
        default="include",
        help="how to handle somatic mutations in "
        "neoepitope enumeration; documentation online for more information",
    )
    call_parser.add_argument(
        "-b",
        "--build",
        type=str,
        required=False,
        help="which default genome build to use (hg19 or GRCh38); "
        "must have used download.py script to install these",
    )
    call_parser.add_argument(
        "-i",
        "--isolate",
        required=False,
        action="store_true",
        help="isolate mutations - do not use phasing information to "
        "combine nearby mutations in the same neoepitope",
    )
    call_parser.add_argument(
        "--nmd",
        required=False,
        action="store_true",
        default=False,
        help="enumerate neoepitopes from nonsense mediated decay transcripts",
    )
    call_parser.add_argument(
        "--pp",
        required=False,
        action="store_true",
        default=False,
        help="enumerate neoepitopes from polymorphic pseudogene transcripts",
    )
    call_parser.add_argument(
        "--igv",
        required=False,
        action="store_true",
        default=False,
        help="enumerate neoepitopes IGV transcripts",
    )
    call_parser.add_argument(
        "--trv",
        required=False,
        action="store_true",
        default=False,
        help="enumerate neoepitopes from TRV transcripts",
    )
    call_parser.add_argument(
        "--allow-nonstart",
        required=False,
        action="store_true",
        help="enumerate neoepitopes from transcripts without annotated start codons",
    )
    call_parser.add_argument(
        "--allow-nonstop",
        required=False,
        action="store_true",
        help="enumerate neoepitopes from transcripts without annotated stop codons",
    )
    args = parser.parse_args()
    if args.subparser_name == "download":
        from .download import NeoepiscopeDownloader

        downloader = NeoepiscopeDownloader()
        downloader.run()
    elif args.subparser_name == "index":
        cds_dict = gtf_to_cds(args.gtf, args.dicts)
        tree = cds_to_tree(cds_dict, args.dicts)
    elif args.subparser_name == "swap":
        adjust_tumor_column(args.input, args.output)
    elif args.subparser_name == "merge":
        combine_vcf(args.germline, args.somatic, outfile=args.output)
    elif args.subparser_name == "prep":
        if args.hapcut2_output:
            prep_hapcut_output(args.output, args.hapcut2_output, args.vcf)
        else:
            prep_hapcut_output(args.output, None, args.vcf)
    elif args.subparser_name == "call":
        # Check that output options are compatible
        if args.fasta and args.output == "-":
            sys.exit(
                "Cannot write fasta results when writing output to standard out; "
                "please specify an output file using the -o/--output option when "
                "using the -f/--fasta flag"
            )
        # Load pickled dictionaries and prepare bowtie index
        if args.build is not None:
            if (
                args.build == "GRCh38"
                and paths.gencode_v27 is not None
                and paths.bowtie_grch38 is not None
            ):
                with open(
                    os.path.join(paths.gencode_v27, "intervals_to_transcript.pickle"),
                    "rb",
                ) as interval_stream:
                    interval_dict = pickle.load(interval_stream)
                with open(
                    os.path.join(paths.gencode_v27, "transcript_to_CDS.pickle"), "rb"
                ) as cds_stream:
                    cds_dict = pickle.load(cds_stream)
                reference_index = bowtie_index.BowtieIndexReference(paths.bowtie_grch38)
            elif (
                args.build == "hg19"
                and paths.gencode_v19 is not None
                and paths.bowtie_hg19 is not None
            ):
                with open(
                    os.path.join(paths.gencode_v19, "intervals_to_transcript.pickle"),
                    "rb",
                ) as interval_stream:
                    interval_dict = pickle.load(interval_stream)
                with open(
                    os.path.join(paths.gencode_v19, "transcript_to_CDS.pickle"), "rb"
                ) as cds_stream:
                    cds_dict = pickle.load(cds_stream)
                reference_index = bowtie_index.BowtieIndexReference(paths.bowtie_hg19)
            else:
                raise RuntimeError(
                    "".join(
                        [
                            args.build,
                            " is not an available genome build. Please "
                            "check that you have run neoepiscope download and are "
                            "using 'hg19' or 'GRCh38' for this argument.",
                        ]
                    )
                )
        else:
            if args.bowtie_index is not None and args.dicts is not None:
                intervals_path = os.path.join(
                    args.dicts, "intervals_to_transcript.pickle"
                )
                if os.path.isfile(intervals_dict):
                    with open(intervals_path, "rb") as interval_stream:
                        interval_dict = pickle.load(interval_stream)
                else:
                    raise RuntimeError(
                        "".join(
                            [
                                "Cannot find ",
                                intervals_path,
                                "; have you indexed your GTF with neoepiscope index?",
                            ]
                        )
                    )
                cds_path = os.path.join(args.dicts, "transcript_to_CDS.pickle")
                if os.path.isfile(cds_path):
                    with open(cds_path, "rb") as cds_stream:
                        cds_dict = pickle.load(cds_stream)
                else:
                    raise RuntimeError(
                        "".join(
                            [
                                "Cannot find ",
                                cds_path,
                                "; have you indexed your GTF with neoepiscope index?",
                            ]
                        )
                    )
                bowtie_files = [
                    "".join([args.bowtie_index, ".", x, ".ebwt"]) for x in range(1, 5)
                ]
                if list(set([os.path.isfile(x) for x in bowtie_files])) == [True]:
                    reference_index = bowtie_index.BowtieIndexReference(
                        args.bowtie_index
                    )
                else:
                    raise RuntimeError("Cannot find specified bowtie index")
            else:
                raise RuntimeError(
                    "User must specify either --build OR "
                    "--bowtie_index and --dicts options"
                )
        # Check affinity predictor
        tool_dict = {}
        if args.no_affinity:
            args.affinity_predictor = None
        if args.affinity_predictor is not None:
            if len(args.affinity_predictor) > 1:
                args.affinity_predictor.remove(["mhcflurry", "1", "affinity,rank"])
            for tool in args.affinity_predictor:
                program = tool[0]
                version = tool[1]
                scoring = tool[2].split(",")
                if "mhcflurry" in program.lower():
                    if version == "1" and "mhcflurry1" not in tool_dict:
                        program = "mhcflurry-predict"
                        acceptable_scoring = ["rank", "affinity", "high", "low"]
                        for method in scoring:
                            if method not in acceptable_scoring:
                                warnings.warn(
                                    " ".join([method, "not compatible with mhcflurry"]),
                                    Warning,
                                )
                                scoring.remove(method)
                        if len(scoring) > 0:
                            tool_dict["mhcflurry1"] = [program, sorted(scoring)]
                    elif "mhcflurry1" in tool_dict:
                        raise RuntimeError(
                            "Conflicting or repetitive installs of mhcflurry given"
                        )
                    else:
                        raise NotImplementedError(
                            " ".join(
                                [
                                    "neoepiscope does not support version",
                                    version,
                                    "of mhcflurry",
                                ]
                            )
                        )
                elif "mhcnuggets" in program.lower():
                    if version == "2" and "mhcnuggets2" not in tool_dict:
                        program = "NA"
                        acceptable_scoring = ["affinity"]
                        for method in scoring:
                            if method not in acceptable_scoring:
                                warnings.warn(
                                    " ".join(
                                        [method, "not compatible with mhcnuggets"]
                                    ),
                                    Warning,
                                )
                                scoring.remove(method)
                        if len(scoring) > 0:
                            tool_dict["mhcnuggets2"] = [program, sorted(scoring)]
                    elif "mhcnuggets2" in tool_dict:
                        raise RuntimeError(
                            "Conflicting or repetitive installs of mhcnuggets given"
                        )
                    else:
                        raise NotImplementedError(
                            " ".join(
                                [
                                    "neoepiscope does not support version",
                                    version,
                                    "of mhcnuggets",
                                ]
                            )
                        )
                elif "netMHCIIpan" in program:
                    if version == "3" and "netMHCIIpan3" not in tool_dict:
                        program = paths.netMHCIIpan3
                        if program is None:
                            program = which("netMHCIIpan3")
                        else:
                            program = which(program)
                        if program is None:
                            warnings.warn(
                                " ".join(
                                    ["No valid install of", "netMHCIIpan available"]
                                ),
                                Warning,
                            )
                            continue
                        acceptable_scoring = ["rank", "affinity"]
                        for method in scoring:
                            if method not in acceptable_scoring:
                                warnings.warn(
                                    " ".join(
                                        [method, "not compatible with netMHCIIpan"]
                                    ),
                                    Warning,
                                )
                                scoring.remove(method)
                        if len(scoring) > 0:
                            tool_dict["netMHCIIpan3"] = [program, sorted(scoring)]
                    elif "netMHCIIpan3" in tool_dict:
                        raise RuntimeError(
                            "Conflicting or repetitive installs of netMHCIIpan given"
                        )
                    else:
                        raise NotImplementedError(
                            " ".join(
                                [
                                    "neoepiscope does not support version",
                                    version,
                                    "of netMHCIIpan",
                                ]
                            )
                        )
                elif "netMHCpan" in program:
                    if ("netMHCpan3" not in tool_dict and version == "3") or (
                        "netMHCpan4" not in tool_dict and version == "4"
                    ):
                        if version == "3":
                            program = paths.netMHCpan3
                            if program is None:
                                program = which("netMHCpan3")
                            else:
                                program = which(program)
                        elif version == "4":
                            program = paths.netMHCpan4
                            if program is None:
                                program = which("netMHCpan4")
                            else:
                                program = which(program)
                        if program is None:
                            warnings.warn(
                                " ".join(
                                    ["No valid install of ", "netMHCIIpan available"]
                                ),
                                Warning,
                            )
                            continue
                        if program is None:
                            warnings.warn(
                                " ".join(
                                    [
                                        "No valid install of",
                                        "netMHCpan version",
                                        version,
                                        "available",
                                    ]
                                ),
                                Warning,
                            )
                            continue
                        acceptable_scoring = ["rank", "affinity"]
                        for method in scoring:
                            if method not in acceptable_scoring:
                                warnings.warn(
                                    " ".join([method, "not compatible with netMHCpan"]),
                                    Warning,
                                )
                                scoring.remove(method)
                        if len(scoring) > 0:
                            if version == "3":
                                name = "netMHCpan3"
                            elif version == "4":
                                name = "netMHCpan4"
                            tool_dict[name] = [program, sorted(scoring)]
                    elif ("netMHCpan3" in tool_dict and version == "3") or (
                        "netMHCpan4" in tool_dict and version == "4"
                    ):
                        raise RuntimeError(
                            "Conflicting or repetitive installs of netMHCpan given"
                        )
                    else:
                        raise NotImplementedError(
                            " ".join(
                                [
                                    "neoepiscope does not support version",
                                    version,
                                    "of netMHCpan",
                                ]
                            )
                        )
                else:
                    raise NotImplementedError(
                        " ".join(
                            [
                                "neoepiscope does not support",
                                program,
                                "for binding predictions",
                            ]
                        )
                    )
        if not tool_dict:
            warnings.warn(
                "No binding prediction tools specified; "
                "will proceed without binding predictions",
                Warning,
            )
            hla_alleles = []
        else:
            if args.alleles:
                hla_alleles = sorted(args.alleles.split(","))
            else:
                raise RuntimeError(
                    "To perform binding affinity predictions, "
                    "user must specify at least one allele "
                    "via the --alleles option"
                )
        # Obtain VAF frequency VCF position
        if args.vcf:
            vaf_pos = get_vaf_pos(args.vcf)
        else:
            vaf_pos = None
        # Obtain peptide sizes for kmerizing peptides
        if "," in args.kmer_size:
            size_list = args.kmer_size.split(",")
            size_list.sort(reverse=True)
            for i in range(0, len(size_list)):
                size_list[i] = int(size_list[i])
        else:
            size_list = [int(args.kmer_size)]
        # Establish handling of ATGs
        if args.upstream_atgs == "none":
            only_novel_upstream = False
            only_downstream = True
            only_reference = False
        elif args.upstream_atgs == "novel":
            only_novel_upstream = True
            only_downstream = False
            only_reference = False
        elif args.upstream_atgs == "all":
            only_novel_upstream = False
            only_downstream = False
            only_reference = False
        elif args.upstream_atgs == "reference":
            only_novel_upstream = False
            only_downstream = False
            only_reference = True
        else:
            raise RuntimeError(
                "--upstream-atgs must be one of "
                '{"novel", "all", "none", "reference"}'
            )
        # Establish handling of germline mutations:
        if args.germline == "background":
            include_germline = 2
        elif args.germline == "include":
            include_germline = 1
        elif args.germline == "exclude":
            include_germline = 0
        else:
            raise RuntimeError(
                "--germline must be one of " '{"background", "include", "exclude"}'
            )
        # Establish handling of somatic mutations:
        if args.somatic == "include":
            include_somatic = 1
        elif args.somatic == "background":
            include_somatic = 2
        elif args.somatic == "exclude":
            include_somatic = 0
        else:
            raise RuntimeError(
                "--somatic must be one of " '{"background", "include", "exclude"}'
            )
        # Warn if somatic and germline are both excluded
        if include_somatic == 0 and include_germline == 0:
            warnings.warn(
                "Germline and somatic mutations are both being "
                "excluded, no epitopes will be returned",
                Warning,
            )
        # Find transcripts that haplotypes overlap
        relevant_transcripts = process_haplotypes(
            args.merged_hapcut2_output, interval_dict, args.isolate
        )
        # Apply mutations to transcripts and get neoepitopes
        neoepitopes, fasta = get_peptides_from_transcripts(
            relevant_transcripts,
            vaf_pos,
            cds_dict,
            only_novel_upstream,
            only_downstream,
            only_reference,
            reference_index,
            size_list,
            args.nmd,
            args.pp,
            args.igv,
            args.trv,
            args.allow_nonstart,
            args.allow_nonstop,
            include_germline,
            include_somatic,
            protein_fasta=args.fasta,
        )
        # If neoepitopes are found, get binding scores and write results
        if len(neoepitopes) > 0:
            full_neoepitopes = gather_binding_scores(
                neoepitopes, tool_dict, hla_alleles
            )
            write_results(args.output, hla_alleles, full_neoepitopes, tool_dict)
            if args.fasta:
                fasta_file = "".join([args.output, ".fasta"])
                with open(fasta_file, "w") as f:
                    for tx in fasta:
                        proteins = sorted(list(fasta[tx]))
                        for i in range(0, len(proteins)):
                            identifier = "".join([">", tx, "_v", str(i)])
                            print(identifier, file=f)
                            print(proteins[i], file=f)
        else:
            sys.exit("No neoepitopes found")
    else:
        parser.print_usage()


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# coding=utf-8
"""
paths.py

Part of neoepiscope
Defines default paths of neoepiscope's dependencies.
None indicates the user didn't install the tool or data.

Licensed under the MIT license.

The MIT License (MIT)
Copyright (c) 2018 Mary A. Wood, Austin Nguyen,
                   Abhinav Nellore, and Reid Thompson

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
gencode_v27 = None
gencode_v19 = None
bowtie_grch38 = None
bowtie_hg19 = None
netMHCIIpan3 = None
netMHCpan3 = None
netMHCpan4 = None
hapcut2_hairs = None
hapcut2 = None



#!/usr/bin/env python
# coding=utf-8

from __future__ import absolute_import, division, print_function
from setuptools import setup, find_packages
from distutils.core import Command
from setuptools.command.install import install

# Borrowed (with revisions) from https://stackoverflow.com/questions/17001010/
# how-to-run-unittest-discover-from-python-setup-py-test/21726329#21726329
class DiscoverTest(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        import os
        import sys
        import unittest

        # get setup.py directory
        setup_file = sys.modules["__main__"].__file__
        test_dir = os.path.join(os.path.abspath(os.path.dirname(setup_file)), "tests")
        # use the default shared TestLoader instance
        test_loader = unittest.defaultTestLoader
        # use the basic test runner that outputs to sys.stderr
        test_runner = unittest.TextTestRunner()
        # automatically discover all tests
        # NOTE: only works for python 2.7 and later
        test_suite = test_loader.discover(test_dir)
        print(test_suite)
        # run the test suite
        test_runner.run(test_suite)


class DownloadDependencies(Command):
    # Wrapper to accommodate old-style class Command
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        from neoepiscope.download import NeoepiscopeDownloader

        downloader = NeoepiscopeDownloader()
        downloader.run()


setup(
    name="neoepiscope",
    version="0.2.1",
    description="comprehensive neoepitope prediction software",
    long_description=(
        "neoepiscope is software for predicting neoepitopes "
        "from DNA sequencing (DNA-seq) data. Where most neoepitope "
        "prediction software confines attention to neoepitopes arising "
        "from at most one somatic mutation, often just an SNV, "
        "neoepiscope uses assembled haplotype output of HapCUT2 "
        "(Edge et al., 2017) to also enumerate neoepitopes arising "
        "from more than one somatic mutation. neoepiscope also takes "
        "into account frameshifting from indels and permits "
        "personalizing the reference transcriptome using germline variants."
    ),
    url="http://github.com/ohsu-comp-bio/neoepiscope",
    download_url="https://github.com/pdxgx/neoepiscope/tarball/v0.2.1",
    author="Mary A. Wood, Austin Nguyen, Abhinav Nellore, Reid F. Thompson",
    author_email="thompsre@ohsu.edu",
    license="MIT",
    packages=["neoepiscope"],
    include_package_data=True,
    package_data={"neoepiscope": ["*.py", "*.pickle"]},
    zip_safe=True,
    install_requires=["intervaltree", "mhcflurry", "mhcnuggets", "numpy==1.14.5"],
    entry_points={"console_scripts": ["neoepiscope=neoepiscope:main"]},
    cmdclass={"download": DownloadDependencies, "test": DiscoverTest},
    keywords=["neoepitope", "neoantigen", "cancer", "immunotherapy"],
    classifiers=[
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Education",
        "Operating System :: MacOS",
        "Operating System :: Unix",
        "Topic :: Scientific/Engineering :: Medical Science Apps.",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
)
